﻿Type=Class
Version=3
@EndOfDesignText@
'Class module
Sub Class_Globals
  Type FTPContRec (prt, typ As Int, con As Socket, opn As Boolean, _
    ist As InputStream, ost As OutputStream, ast As AsyncStreams, usr, pwd, str As String)
  Type FTPDataRec (prt As Int, con As Socket, opn, abo As Boolean, _
    ist, fist As InputStream, ost, fost As OutputStream, ast As AsyncStreams, fname, cmd As String)
  ' prt - server socket port
  ' typ - 0-binary mode (def), 1-ASCII mode
  ' con - socket connection
  ' opn - active connection
  ' abo - abort request received
  ' ist, ost - connection input stream, output stream
  ' fist, fost - file in/out stream
  ' ast - async stream for data
  ' usr, pwd - ContSoc: user name and password of connected user
  ' fname - File name for retr stor dele
  ' str - partially received command string buffer
  ' cmd - command string for data timer

  Dim ContSvrSoc As ServerSocket ' Control socket server (listener)
  Dim DataSvrSoc As ServerSocket ' Data socket server (listener)
  Dim ContSoc As FTPContRec
  Dim DataSoc As FTPDataRec
  
  ContSoc.Initialize ' init types
  DataSoc.Initialize
  
  Dim tmrCont As Timer ' FTP control connection timer
  tmrCont.Initialize("tmrCont", 100)

  Dim tmrData As Timer ' Data connection timer
  tmrData.Initialize("tmrData", 100)
  
  Dim pnlFTP As Panel
  Dim HomeDir As String
  
  Dim lstDebug As List
End Sub

'Initializes the object. You can add parameters to this method if needed.
Public Sub Initialize(pnl As Panel, hdir As String)
  ContSoc.prt = 2121
  ContSoc.typ = 0

  DataSoc.prt = 2122
  
  pnlFTP = pnl
  HomeDir = hdir
  
  Dim lbl As Label
  lbl.Initialize("")
  pnlFTP.AddView(lbl, 0, 0, pnlFTP.Width, pnlFTP.Height)
  
  lbl.Color = Colors.ARGB(255, 0, 0, 32)
  lbl.TextColor = Colors.White
  lbl.TextSize = 16
  lbl.Gravity = Gravity.TOP + Gravity.CENTER_HORIZONTAL

  Dim btn As Button
  btn.Initialize("CloseBtn")
  pnlFTP.AddView(btn, 0, 0, 140dip, 50dip)
  btn.Text = "Close FTP"
  
  btn.Left = (pnlFTP.Width - btn.Width) / 2
  btn.Top = pnlFTP.Height - btn.Height - 2dip
  
  lbl.Height = pnlFTP.Height - btn.Height - 8dip
  
  pnlFTP.Enabled = False
  pnlFTP.Visible = False
  pnlFTP.SendToBack
  
  lstDebug.Initialize ' *** debug
End Sub

Public Sub OpenSvr
  Dim lbl As Label
  
  ContSvrSoc.Initialize(ContSoc.prt, "ContSvrSoc")
  ContSoc.opn = False
  
  DataSvrSoc.Initialize(DataSoc.prt, "DataSvrSoc")
  DataSoc.opn = False
  
  lbl = pnlFTP.GetView(0)

  lbl.Text = Sprintf("FTP Server IP: %d\r\n", Array As Object(ContSvrSoc.GetMyIP))
  lbl.Text = lbl.Text & Sprintf("FTP Port: %d\r\n\r\n", Array As Object(ContSoc.prt))
  
  ContSvrSoc.Listen

  lbl.Text = lbl.Text & "Listening ..." & CRLF & CRLF
  
  pnlFTP.Enabled = True
  pnlFTP.Visible = True
  pnlFTP.BringToFront
End Sub

Sub CloseBtn_Click
  CloseSvr
  
  pnlFTP.Enabled = False
  pnlFTP.Visible = False
  pnlFTP.SendToBack
  
  CallSubDelayed(Main, "ShowFiles")
End Sub

Public Sub CloseSvr
  tmrCont.Enabled = False
  tmrData.Enabled = False
    
  CloseContSoc
  ContSvrSoc.close 'stop listening
    
  CloseDataSoc
  DataSvrSoc.close 'stop listening
End Sub

Sub FTPCmdOutput(sa() As String)
  Dim s As String
  Dim lbl As Label
  
  s = sa(0)
  
  If (sa.Length = 2) Then
    If (s.ToLowerCase = "pass") Then sa(1) = "(password not shown)"
    s = s & " " & sa(1)
  End If
  
  lbl = pnlFTP.GetView(0)
  
  lbl.Text = lbl.Text & s & CRLF  
End Sub

Sub tmrCont_Tick
  tmrCont.Enabled = False

  If (ContSoc.opn = False) Then
    CloseContSoc
    Return
  End If
  
  If (ContSoc.str.EndsWith(CRLF) = True) Then
    Dim s, sa() As String
      
    sa = Regex.Split(" ", ContSoc.str.trim)
    ContSoc.str = ""
        
'    Mandatory commands:
'    
'    ABOR, ACCT, ALLO, APPE, CWD, DELE, HELP, LIST, MODE, NLST, NOOP,
'    PASS, PASV, PORT, QUIT, REIN, REST, RETR, RNFR, RNTO, SITE, STAT,
'    STOR, STRU, TYPE, USER

    FTPCmdOutput(sa)

    Select (sa(0).ToLowerCase)
      Case "abor"
        If (DataSoc.opn = True) Then
          DataSoc.abo = True
        Else
          s = Sprintf("%d Aborted OK\r\n", Array As Object(200))
        End If
            
      Case "cwd"
        ' return success but do nothing
        s = Sprintf("%d Command successful (%s)\r\n", Array As Object(250, HomeDir))

      Case "dele"
        If (File.Exists(HomeDir, sa(1)) = True) Then
          Try
            File.Delete(HomeDir, sa(1))
            s = Sprintf("%d File %s successfully deleted\r\n", Array As Object(250, sa(1)))
          Catch
            s = Sprintf("%d Error deleting %s\r\n", Array As Object(550, sa(1)))
          End Try                
        Else
          s = Sprintf("%d File %s not found\r\n", Array As Object(550, sa(1)))
        End If
            
      Case "list"
        DataSoc.cmd = "list"
        s = Sprintf("%d Directory Listing for %s\r\n", Array As Object(150, HomeDir))

      Case "noop"
        s = Sprintf("%d OK\r\n", Array As Object(200))

      Case "pass"
        ContSoc.pwd = sa(1)
        s = Sprintf("%d Password accepted\r\n", Array As Object(230))

      Case "pasv"
        Dim i1, i2 As Int
        Dim ipa() As String
            
        i1 = DataSoc.prt / 256
        i2 = DataSoc.prt Mod 256
                        
        ipa = Regex.Split("\.", DataSvrSoc.GetMyIP)

        s = Sprintf("%s,%s,%s,%s,%d,%d", Array As Object(ipa(0), ipa(1), ipa(2), ipa(3), i1, i2))
        s = Sprintf("%d Entering passive mode (%s)\r\n", Array As Object(227, s))
            
        DataSoc.cmd = "pasv"
        DataSvrSoc.Listen ' accept data socket connection

      Case "pwd"
        s = QUOTE
        s = Sprintf("%d %s%s%s\r\n", Array As Object(257, s, HomeDir, s))

      Case "quit"
        s = Sprintf("%d Connection closed OK\r\n", Array As Object(200))
        SendContStr(s)
            
        CloseContSoc
            
        If (DataSoc.opn = False) Then CloseDataSoc
           
        Return ' tmrCont disabled

      Case "retr"
        DataSoc.fname = sa(1)
        DataSoc.cmd = "retr"
        s = "" ' response sent from tmrData, tmrData already enabled
              
      Case "stor"
        DataSoc.fname = sa(1)
        DataSoc.cmd = "stor"
        s = "" ' response sent from tmrData, tmrData already enabled
              
      Case "type"
        If (sa(1).ToLowerCase.CharAt(0) = "a") Then
          ContSoc.typ = 1
          s = Sprintf("%d Switching to ASCII mode\r\n", Array As Object(200))
        Else
          ContSoc.typ = 0
          s = Sprintf("%d Switching to binary mode\r\n", Array As Object(200))
        End If

      Case "user"
        ContSoc.usr = sa(1)
        s = Sprintf("%d Login as %s OK. Send password\r\n", Array As Object(331, sa(1)))

      Case Else
        s = Sprintf("%d Command not implemented\r\n", Array As Object(502))
    End Select
        
    If (s <> "") Then SendContStr(s)
  End If

  tmrCont.Enabled = True
End Sub

Sub SendContStr(str As String)
  Dim obuf(str.Length) As Byte
  Dim bc As ByteConverter
      
  obuf = bc.StringToBytes(str, "UTF8")
  ContSoc.ast.Write(obuf)
End Sub

Sub CloseContSoc
  ContSoc.opn = False
  If (ContSoc.ist.IsInitialized) Then ContSoc.ist.close
  If (ContSoc.ost.IsInitialized) Then ContSoc.ost.close
  ContSoc.con.close
End Sub

Sub ContSvrSoc_NewConnection (Success As Boolean, NewSocket As Socket)
  If (Success = True) Then
    Dim s As String

    If (ContSoc.opn = True) Then
      Log("FTP Socket Busy Error")
      Return
    End If
    
    ContSoc.opn = True
    ContSoc.con = NewSocket
    ContSoc.ist = ContSoc.con.InputStream
    ContSoc.ost = ContSoc.con.OutputStream
    ContSoc.str = ""
	
	ContSoc.ast.Initialize(ContSoc.ist, ContSoc.ost, "AContStm")
    
    Log("Control Socket Connected")
    
    s = Sprintf("%d Welcome to MyFtp\r\n", Array As Object(220))
    
    SendContStr(s)
    
    tmrCont.Enabled = True
  Else
    Log("Control Socket Success Error: " & LastException.Message)
  End If
  
  ContSvrSoc.Listen 'Continue listening to new incoming connections
End Sub

Sub AContStm_NewData(Buffer() As Byte)
  Dim s As String
  Dim bc As ByteConverter
      
  s = bc.StringFromBytes(Buffer, "UTF8")
  ContSoc.str = ContSoc.str & s
End Sub

Sub AContStm_Terminated
  DataSoc.opn = False
End Sub

Sub AContStm_Error
  Dim s As String
  
  s = LastException.Message
  Log("AContStm error: " & s)

  CloseContSoc
End Sub

Sub tmrData_Tick
  Dim i As Int
  Dim s As String
  
  tmrData.Enabled = False

  If (DataSoc.abo = True) Then
    s = Sprintf("%d Connection closed. Requested action aborted\r\n", Array As Object(451))
    SendContStr(s)

    CloseDataSoc
    Return ' timer will remain disabled
  End If
  
  Select (DataSoc.cmd)
    Case "list"
      Dim dt As Long
      Dim sa(), sb(7) As String
      Dim lst As List
      Dim per As Period
      Dim fs As MLfiles

        
      ' d---rwxr-x system sdcard_rw 2011-11-13 12:59 LOST.DIR
      ' permissions and type, the first Char indicates a dir, link, or File (d, l, or -).
      ' owner, group, size, date, time, File name

      lst = fs.dirEntries(HomeDir)
      
      If (lst.IsInitialized = True) Then
        Log(lst)

        ' convert to format
        ' permissions, number of hard links, owner, group, size, date(3), and filename
          
       DataSoc.ast.Initialize(DataSoc.ist, DataSoc.ost, "ADataStm")

       For i=0 To lst.Size-1
          s = lst.Get(i)
          sa = Regex.Split(" ", s)
          
          If (sa.Length <> 7) Then Continue
          If (sa(0).CharAt(0) <> "-") Then Continue ' check for file (not dir or link)
           
          sb(0) = sa(0) ' permissions
          sb(1) = 0 ' num hard links
          sb(2) = sa(1) ' owner
          sb(3) = sa(2) ' group
          sb(4) = sa(3) ' size
            
          DateTime.DateFormat = "y-M-d"
          DateTime.TimeFormat = "H:m"
          dt = DateTime.DateTimeParse(sa(4), sa(5)) ' date, time
          per = DateUtils.PeriodBetween(dt, DateTime.Now)
            
          If ((per.Months < 6) OR (dt > DateTime.Now)) Then
            sb(5) = FormatDateTime("MMM dd HH:mm", dt)
          Else
            sb(5) = FormatDateTime("MMM dd yyyy", dt)
          End If
            
          sb(6) = sa(6)
            
          s = Sprintf("%s %s %s %s %s %s %s\r\n", _
            Array As Object(sb(0), sb(1), sb(2), sb(3), sb(4), sb(5), sb(6)))
              
          Log(s)
            
          SendDataStr(s)
        Next
      End If

      DataSoc.cmd = "list2"
        
    Case "list2"
      s = Sprintf("%d Directory send OK\r\n", Array As Object(226))
      SendContStr(s)
        
      CloseDataSoc
      Return ' timer will remain disabled
        
    Case "retr"
      Try
        DataSoc.fist = File.OpenInput(HomeDir, DataSoc.fname)
        
        s = Sprintf("%d File status OK. Sending %s\r\n", Array As Object(150, DataSoc.fname))
        SendContStr(s)

        DataSoc.ast.Initialize(DataSoc.ist, DataSoc.ost, "ADataStm")
        
        DataSoc.cmd = "retr2"
      Catch
        s = Sprintf("%d Requested action aborted\r\n", Array As Object(451))
        SendContStr(s)

        CloseDataSoc
        Return ' timer will remain disabled
      End Try
        
    Case "retr2"
      If (DataSoc.fist.BytesAvailable > 0) Then
        Dim buf(10000) As Byte
          
        Try
          i = DataSoc.fist.ReadBytes(buf, 0, buf.Length)
          DataSoc.ast.Write2(buf, 0, i)
        Catch
          s = Sprintf("%d Requested action aborted\r\n", Array As Object(451))
          SendContStr(s)
            
          DataSoc.fist.close
          CloseDataSoc
          Return ' timer will remain disabled
        End Try
      End If
        
      If (DataSoc.fist.BytesAvailable <= 0) Then
        s = Sprintf("%d Transfer complete\r\n", Array As Object(226))
        SendContStr(s)

        DataSoc.fist.close
        CloseDataSoc
        Return ' timer will remain disabled
      End If
        
    Case "stor"
      Try
        DataSoc.fost = File.OpenOutput(HomeDir, DataSoc.fname, False)
        
        s = Sprintf("%d File status OK. Waiting for %s\r\n", Array As Object(150, DataSoc.fname))
        SendContStr(s)

        DataSoc.cmd = "stor2"
		
		lstDebug.Clear ' *** debug

        DataSoc.ast.Initialize(DataSoc.ist, DataSoc.ost, "ADataStm")
          
      Catch
        s = Sprintf("%d Requested action aborted\r\n", Array As Object(451))
        SendContStr(s)

        CloseDataSoc
        Return ' timer will remain disabled
      End Try
        
    Case "stor2"
      If (DataSoc.opn = False) Then
        s = Sprintf("%d Transfer complete\r\n", Array As Object(226))
        SendContStr(s)

        DataSoc.fost.close
        CloseDataSoc
		
        ' *** start debug
        Dim dbn As Int = 0
        Dim dbv As Int = 0
		
        Log("STOR debug buffer xfers")
		
        For dbi=0 To lstDebug.Size-1
          dbv = lstDebug.Get(dbi)
          dbn = dbn + dbv
		  
'          Log("  " & dbv)
        Next
		
        Log("STOR buffer total=" & dbn)
        ' *** end debug

        Return ' timer will remain disabled
      End If
          
  End Select
    
  tmrData.Enabled = True
End Sub

Sub SendDataStr(str As String)
  Dim obuf(str.Length) As Byte
  Dim bc As ByteConverter
      
  obuf = bc.StringToBytes(str, "UTF8")
  DataSoc.ast.Write(obuf)
End Sub

Sub CloseDataSoc
  DataSoc.opn = False
  If (DataSoc.ist.IsInitialized) Then DataSoc.ist.close
  If (DataSoc.ost.IsInitialized) Then DataSoc.ost.close
  DataSoc.con.close

  If (DataSoc.fist.IsInitialized = True) Then DataSoc.fist.close
  If (DataSoc.fost.IsInitialized = True) Then DataSoc.fost.close
End Sub

Sub DataSvrSoc_NewConnection (Success As Boolean, NewSocket As Socket)
  If (Success = True) Then
    Dim s As String
    Dim bc As ByteConverter

    If (DataSoc.opn = True) Then
      Log("Data Socket Busy Error")
      Return
    End If
    
    DataSoc.opn = True
    DataSoc.abo = False
    DataSoc.cmd = ""
    DataSoc.con = NewSocket
    DataSoc.ist = DataSoc.con.InputStream
    DataSoc.ost = DataSoc.con.OutputStream
    DataSoc.fist = Null
    DataSoc.fost = Null
    
    Log("Data Socket Connected")
    
    tmrData.Enabled = True
  Else
    Log("Data Socket Success Error: " & LastException.Message)
  End If
  
  ' implicit stop listening ?
End Sub

Sub ADataStm_NewData(Buffer() As Byte)
  If (DataSoc.cmd = "stor2") Then
    DataSoc.fost.WriteBytes(Buffer, 0, Buffer.Length)
    lstDebug.Add(Buffer.Length) ' *** debug
  End If
End Sub

Sub ADataStm_Terminated
  DataSoc.opn = False
End Sub

Sub ADataStm_Error
  Dim s As String
  
  s = LastException.Message
  Log("ADataStm error: " & s)

  CloseDataSoc
End Sub

Sub Sprintf(fmt As String, arg() As Object) As String
  Dim ai, fi, i, j As Int
  Dim exp, wid As Int
  Dim stmp, ptmp, sres, C, S, t As String
  Dim bdone As Boolean
  
  ai = 0
  fi = 0
  stmp = ""
  sres = ""
  
  Do While (fi < fmt.Length)
    C = fmt.CharAt(fi)
    fi = fi + 1
    
    If (C = "%") Then
      stmp = ""
      ptmp = ""
      bdone = False
      
      Do While ((fi < fmt.Length) AND (bdone = False))
        C = fmt.CharAt(fi)
        fi = fi + 1
        
        If (C = "%") Then
          sres = sres & C
          bdone = True
        Else If ((IsNumber(C) = True) OR (C = ".")) Then
          ptmp = ptmp & C
          bdone = False
        Else If (C = "s") Then
          stmp = arg(ai)
          ai = ai + 1
          sres = sres & stmp
          bdone = True
        Else If (C = "d") Then
          If (ptmp.Length > 0) Then
            t = " "
            If (ptmp.CharAt(0) = "0") Then
              t = "0"
              ptmp = ptmp.SubString(1)
            End If
          End If
          
          If (ptmp.Length > 0) Then wid = ptmp Else wid = 0
          
          stmp = arg(ai)
          
          Do While (stmp.Length < wid)
            stmp = t & stmp
          Loop
          
          ai = ai + 1
          sres = sres & stmp
          bdone = True
        Else If (C = "f") Then
          If (ptmp.Length > 0) Then
            i = ptmp.IndexOf(".")
            If (i >= 0) Then
              wid = ptmp.SubString2(0, i)
              exp = ptmp.SubString(i+1)
            Else
              wid = ptmp
              exp = 0
            End If
            
            stmp = NumberFormat(arg(ai), wid, exp)
          Else
            stmp = arg(ai)
          End If

          ai = ai + 1
          sres = sres & stmp
          bdone = True
        Else
          ai = ai + 1
        End If
        
      Loop
    Else
      sres = sres & C
    End If
  Loop
  
  Return(TRepl(sres))
End Sub

Sub TRepl(txt As String) As String
  Dim s As String
  
  s = txt
  
  s = s.Replace("\n", Chr(10))
  s = s.Replace("\r", Chr(13))
  s = s.Replace("\t", Chr(9))

  Return(s)
End Sub

Sub FormatDateTime(fmt As String, dt As Long) As String
  Dim i, j As Int
  Dim s, t, so, MonthNames(12) As String
  
  s = fmt
  so = ""
  MonthNames = Array As String("Jan", "Feb", "Mar", "Apr", _
    "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec")
  
  Do While (s.Length > 0)
    If (s.StartsWith("HH")) Then
      t = NumberFormat2(DateTime.GetHour(dt), 2, 0, 0, False)
      j = 2
    Else If (s.StartsWith("H")) Then
      t = NumberFormat2(DateTime.GetHour(dt), 1, 0, 0, False)
      j = 1
    Else If (s.StartsWith("hh")) Then
      i = DateTime.GetHour(dt)
      If (i > 12) Then i = i - 12
      t = NumberFormat2(i, 2, 0, 0, False)
      j = 2
    Else If (s.StartsWith("h")) Then
      i = DateTime.GetHour(dt)
      If (i > 12) Then i = i - 12
      t = NumberFormat2(i, 1, 0, 0, False)
      j = 1
    Else If (s.StartsWith("mm")) Then
      t = NumberFormat2(DateTime.GetMinute(dt), 2, 0, 0, False)
      j = 2
    Else If (s.StartsWith("m")) Then
      t = NumberFormat2(DateTime.GetMinute(dt), 1, 0, 0, False)
      j = 1
    Else If (s.StartsWith("ss")) Then
      t = NumberFormat2(DateTime.GetSecond(dt), 2, 0, 0, False)
      j = 2
    Else If (s.StartsWith("yyyy")) Then
      t = NumberFormat2(DateTime.GetYear(dt), 1, 0, 0, False)
      j = 4
    Else If (s.StartsWith("yy")) Then
      t = NumberFormat2(DateTime.GetYear(dt), 1, 0, 0, False)
      i = Max(t.Length - 2, 0)
      t = t.SubString(i)
      j = 2
    Else If (s.StartsWith("MMM")) Then
      i = DateTime.GetMonth(dt)
      t = MonthNames(i - 1)
      j = 3
    Else If (s.StartsWith("MM")) Then
      t = NumberFormat2(DateTime.GetMonth(dt), 2, 0, 0, False)
      j = 2
    Else If (s.StartsWith("M")) Then
      t = NumberFormat2(DateTime.GetMonth(dt), 1, 0, 0, False)
      j = 1
    Else If (s.StartsWith("dd")) Then
      t = NumberFormat2(DateTime.GetDayOfMonth(dt), 2, 0, 0, False)
      j = 2
    Else If (s.StartsWith("d")) Then
      t = NumberFormat2(DateTime.GetDayOfMonth(dt), 1, 0, 0, False)
      j = 1
    Else
      t = s.SubString2(0, 1)
      j = 1
    End If
    
    so = so & t
    s = s.SubString(j)
  Loop
  
  Return(so)
End Sub
